# -*- coding: utf-8 -*-

from qtalchemy import MapperMixin, user_message, LayoutLayout, LayoutWidget, Message, CommandEvent
from PyQt4 import QtCore, QtGui
from sqlalchemy.orm import mapper, create_session, relation, object_session, Query
import traceback
import sys

def MessageButtonsToQt(flags):
    result = QtGui.QMessageBox.NoButton
    for x in "Ok Cancel Yes No".split(' '):
        if flags & getattr(Message, x):
            result |= getattr(QtGui.QMessageBox, x)
    return result

def QtToMessageButtons(flags):
    result = 0
    for x in "Ok Cancel Yes No".split(' '):
        if flags & getattr(QtGui.QMessageBox, x):
            result |= getattr(Message, x)
    return result

def appMessage(parent, flags, msg):
    """
    Display a Qt message box.
    """
    QtGui.QApplication.setOverrideCursor(QtGui.QCursor(QtCore.Qt.ArrowCursor))

    try:
        msgBox = QtGui.QMessageBox(parent)
        msgBox.setWindowTitle(QtGui.QApplication.applicationName())
        #if icon is not None:
        #    msgBox.setIcon( icon )
        msgBox.setText( msg )
        
        msgBox.setStandardButtons(MessageButtonsToQt(flags))
        return QtToMessageButtons(msgBox.exec_())
    finally:
        QtGui.QApplication.restoreOverrideCursor()

class SessionMessageBox(object):
    """
    This class is a callable to hook to a session message handler to display 
    messages for this session.
    """
    def __init__(self, parent=None):
        self.parent = parent

    def __call__(self, session, instance, flags, msg):
        # TODO:  pass icon
        # TODO:  pass exception info
        # TODO:  pass additional buttons
        # TODO:  allow data entry on instance (??) -- augment the message box with a bound data form
        msgBox = QtGui.QMessageBox(self.parent)
        msgBox.setWindowTitle(QtGui.QApplication.applicationName())
        #if icon is not None:
        #    msgBox.setIcon( icon )
        msgBox.setText( msg )
        
        msgBox.setStandardButtons(MessageButtonsToQt(flags))
        return QtToMessageButtons(msgBox.exec_())

def exception_message_box(e,text,parent=None,icon=None):
    """
    Extract exception information and display a message box with exception details in 
    a details extension of the message.
    """
    msg = QtGui.QMessageBox(parent)
    msg.setWindowTitle(QtGui.QApplication.applicationName())
    if icon is not None:
        msg.setIcon( icon )
    msg.setText( text )
    msg.setInformativeText( user_message(e) )
    exc_type, exc_value, exc_traceback = sys.exc_info()
    msg.setDetailedText( ''.join(traceback.format_exception(exc_type, exc_value, exc_traceback)) )
    msg.setStandardButtons( QtGui.QMessageBox.Ok )
    return msg.exec_()

def messaged_commit(session, parent):
    """
    This helper function displays a message to the user if the commit on the session fails.
    
    The message text is given by the user_message function.
    """
    try:
        session.commit()
    except Exception, e:
        exception_message_box(e,"There was an error saving the data.",parent=parent,icon=QtGui.QMessageBox.Warning)

class MapperDialog(QtGui.QDialog, MapperMixin):
    def __init__(self,parent):
        QtGui.QDialog.__init__(self,parent)

    def accept(self):
        self.submit()
        QtGui.QDialog.accept(self)

class BoundDialog(MapperDialog):
    """
    The BoundDialog class pulls together the :class:`MapperMixin` and 
    sqlalchemy session management.  There are functions to load and save a 
    data form.
    
    The typical set-up code for a BoundDialog object has the following in 
    __init__:

      - a call to setDataReader giving the table and key information to get a 
        record to display on screen
      - code setting up the widgets on the screen (particularly with calls to 
        the inherited :func:`MapperMixin.mapClass`)
      - a final call to self.readData with the initial record to display

    An application utilizing methods :func:`preCommandSave`, :func:`refresh`, 
    and :func:`readData` may re-implement :func:`load`, but it is recommended 
    to call add screen loaders with :func:`addLoader`.
    """
    def __init__(self, parent, session=None, flush=True):
        MapperDialog.__init__(self,parent)
        self.flush_on_close = flush
        self.session = session
        self.loaders = []

    def setDataReader(self, Session, tableClass, keyAttr):
        """
        :param Session:  Session class returned from sqlalchemy sessionmaker
        :param tableClass:  sqlalchemy mapped class with the root of the 
            data shown on this form
        :param keyAttr:  attribute of primary key in tableClass
        """
        self.Session = Session
        self.tableClass = tableClass
        self.keyAttr = keyAttr
        if self.tableClass is not None and self.keyAttr is not None:
            self.keyColumn = getattr(self.tableClass, self.keyAttr)

    def readData(self, row=None, rowKey=None, load=True):
        """
        If row is specified then row is expected to be an instance of
        self.tableClass.  If neither is specified, then no data is loaded and a
        new class of type self.tableType is created.  One can use either
        parameter for flexibility to specify the data to load by object or by
        primary key.

        :param row: row instance to map from the bound tableClass
        :param rowKey:  primary key value which will be loaded from the database 
            for display
        """
        if row is not None:
            self.main_row = row
            self.session = object_session(row)
        else:
            self.session = self.Session()
            if rowKey is None:
                self.main_row = self.tableClass()
                self.session.add(self.main_row)
            else:
                self.main_row = self.session.query(self.tableClass).filter(self.keyColumn==rowKey).one()

        if load:
            self._load()

    def addLoader(self, callable):
        """
        Add a collable taking the main row and which loads the data from this
        main row displaying it on the screen.
        
        :param callable:  callable taking self.main_row
        """
        self.loaders.append(callable)

    def _load(self):
        for l in self.loaders:
            l(self.main_row)
        self.load()

    def load(self):
        """
        This load method should connect mappers with
        :func:`WidgetAttributeMapper.connect_instance` and connect models with
        :func:`PBTableModel.reset_content_from_list`.  The default
        implementation of load does nothing.
        """
        pass

    def preCommandSave(self, event):
        """
        This function is a recommended slot for the preCommandSave signal of 
        :class:`BoundCommandMenu`.

        Application developers may call this function in tandem with 
        :func:`refresh` in the following way.
        
            >>> event = CommandEvent()
            >>> self.preCommandSave(event)        #doctest: +SKIP
            >>> if not event.aborted:             #doctest: +SKIP
            ...     # take some action with the persisted data
            ...     self.refresh(event)
        """
        try:
            event.savedDataContext = self.main_row
            MapperDialog.submit(self)
            if self.flush_on_close:
                self.flush()
                self.session.close()
        except Exception, e:
            self.session.rollback()
            exception_message_box(e,"There was an error saving data.",icon=QtGui.QMessageBox.Warning,parent=self)
            event.abort()

    def refresh(self, event):
        """
        This function is a recommended slot for the refresh signal of 
        :class:`BoundCommandMenu`.
        
        Note that it calls self.load which the application is expected to 
        provide (there is no default implementation).
        """
        self.readData(rowKey = getattr(event.savedDataContext, self.keyAttr), load=True)

    def commitAndRefresh(self):
        event  = CommandEvent()
        self.preCommandSave(event)
        if not event.aborted:
            self.refresh(event)

    def accept(self):
        event  = CommandEvent()
        self.preCommandSave(event)
        if not event.aborted:
            QtGui.QDialog.accept(self)

    def reject(self):
        MapperDialog.reject(self)
        if self.flush_on_close and self.session is not None:
            self.session.close()

    def flush(self):
        assert self.session is not None, "If flushing changes, we must have a session"
        self.session.commit()

def BoundDialogObject(parent,row):
    b = BoundDialog(parent)
    b.setDataReader(object_session(row), None, None)
    b.setWindowTitle(row.DialogTitle)

    main = QtGui.QVBoxLayout()
    b.setLayout(main)
    grid = LayoutLayout(main,QtGui.QFormLayout())

    m = b.mapClass(type(row))
    for col in row.DialogVisualElements:
        m.addBoundField(grid,col)
    m.connect_instance(row)

    buttonbox = LayoutWidget(main,QtGui.QDialogButtonBox())
    buttonbox.addButton(QtGui.QDialogButtonBox.Ok)
    buttonbox.addButton(QtGui.QDialogButtonBox.Cancel)
    buttonbox.accepted.connect(b.accept)
    buttonbox.rejected.connect(b.reject)

    b.show()
    b.exec_()



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################

from qtalchemy import writeTableColumnGeo

class ViewBase(object):
    """
    This handles functions common to TreeView and TableView.  At this 
    point, it's innards are not well encapsulated, but at least the 
    code is not duplicated.
    """
    def nextIndex(self,index):
        model = self.model()
        if index.column()+1 < model.columnCount(index.parent()):
            return model.index(index.row(),index.column()+1,index.parent())
        elif index.row()+1 < model.rowCount(index.parent()):
            return model.index(index.row()+1,0,index.parent())
        else:
            return model.index(0,0,index.parent())

    def prevIndex(self,index):
        model = self.model()
        if index.column() > 0:
            return model.index(index.row(),index.column()-1,index.parent())
        elif index.row() > 0:
            return model.index(index.row()-1,model.columnCount(index.parent())-1,index.parent())
        else:
            return model.index(index.rowCount(index.parent())-1,model.columnCount(index.parent())-1,index.parent())

    def delKeyPressed(self):
        #print "deleting"
        model = self.model()
        index = self.currentIndex()
        
        # We'll just call removeRows.  If the model doesn't support it, it returns False
        model.removeRows(index.row(),1,index.parent())

    def saveSections(self):
        if self.property("ExtensionId") is not None and self.model() is not None:
            writeTableColumnGeo(self, self.property("ExtensionId"))




# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################

from qtalchemy import *
from PyQt4 import QtCore, QtGui

# patterned after http://labs.qt.nokia.com/2007/06/06/lineedit-with-a-clear-button/

class PBButtonEdit(QtGui.QLineEdit):
    buttonPressed = Signal(name='buttonPressed')

    def __init__(self,parent=None):
        QtGui.QLineEdit.__init__(self,parent)
        self.button = QtGui.QToolButton(self)
        self.button.setCursor(QtCore.Qt.ArrowCursor)
        self.button.setFocusPolicy(QtCore.Qt.NoFocus)
        buttonWidth = self.style().pixelMetric(QtGui.QStyle.PM_ScrollBarExtent)
        #self.setStyleSheet("QLineEdit {{ padding-right: {0}px; }}".format(buttonWidth))

        self.button.clicked.connect(self.buttonPress)

    def resizeEvent(self,event):
        rect = self.rect()
        frameWidth = self.style().pixelMetric(QtGui.QStyle.PM_DefaultFrameWidth)
        buttonWidth = self.style().pixelMetric(QtGui.QStyle.PM_ScrollBarExtent)
        self.button.resize(buttonWidth,rect.height()-2*frameWidth)
        self.button.move(rect.right() - buttonWidth, frameWidth)

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_F4:
            self.buttonPress()
        else:
            QtGui.QLineEdit.keyPressEvent(self, event)

    def buttonPress(self):
        self.buttonPressed.emit()



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
The search dialog is a very high level GUI element of the qtalchemy library.  It utilizes 
the DomainEntity to provide descriptive details of the objects to query in these search dialogs.
"""

from PyQtHelpers import *
from sqlalchemy.orm import mapper, create_session, relation, object_session
import sqlalchemy.sql.expression as expr
from sqlalchemy_helper import *
import sqlalchemy
from qtalchemy.widgets import TableView
from PyQtModels import *

from commands import Command, BoundCommandMenu

class PBSearchableListDialog(QtGui.QDialog):
    """
    This base class gathers a search edit and table together in a single dialog.
    
    This class is now deprecated in favor of :class:`qtalchemy.ext.QueryDataView`.
    """
    def __init__(self, parent=None, extensionId=None):
        QtGui.QDialog.__init__(self, parent, QtCore.Qt.Dialog)

        self.setObjectName(extensionId)

        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        searchLayout = QtGui.QHBoxLayout()
        vbox.addLayout(searchLayout)
        searchLabel = QtGui.QLabel("&Search: ")
        searchLayout.addWidget(searchLabel)
        self.searchEdit = QtGui.QLineEdit()
        self.searchEdit.returnPressed.connect(self.resetSearch)
        searchLabel.setBuddy(self.searchEdit)
        searchLayout.addWidget(self.searchEdit)
        self.searchGo = QtGui.QPushButton("&Refresh")
        searchLayout.addWidget(self.searchGo)
        self.searchGo.clicked.connect(self.resetSearch)
 
        self.table = TableView()
        vbox.addWidget(self.table)
        
        self.setFocusProxy(self.searchEdit)

    def resetSearch(self):
        if self.searchEdit.text() != "":
            like_str = "%%%s%%"%(self.searchEdit.text(),)
            filter_cols = [c.ilike(like_str) for c in self.entity.list_search_columns]

            self.table.model().query = self.base_query.filter(expr.or_(*tuple(filter_cols)))
        else:
            self.table.model().query = self.base_query

        self.table.model().reset_content_from_session()

    def keyPressEvent(self, e):
        if e.key() == QtCore.Qt.Key_Down:
            self.table.setFocus(QtCore.Qt.TabFocusReason)
            e.accept()
        else:
            QtGui.QDialog.keyPressEvent(self, e)

class PBSearchDialog(PBSearchableListDialog):
    """
    This class is intended to be used principally by the PBKeyEdit to provide a
    graphical search dialog which is a little more featureful than can be
    packed into a basic combo box.

    This class is now deprecated in favor of
    :class:`qtalchemy.ext.QueryDataView`.
    """
    def __init__(self, ssrc, entityCls, parent=None):
        extensionId = "%s/Search" % (entityCls.__name__, )

        PBSearchableListDialog.__init__(self, parent, extensionId=extensionId)
        self.setWindowTitle("Search")

        self.ssrc = ssrc
        self.entity = entityCls(ssrc, self)
        self.base_query, converter = self.entity.list_query_converter()

        self.table.setModel(QueryTableModel(self.base_query, ssrc=ssrc, objectConverter=converter), 
                            extensionId=suffixExtId(self, "Table"))

        self.toolbar = LayoutWidget(self.layout(), QtGui.QToolBar(self))
        buttonbox = LayoutWidget(self.layout(),QtGui.QDialogButtonBox())

        self.bindings = self.entity.itemCommands.withView(self.table, bindDefault=False)
        self.bindings.fillToolbar(self.toolbar)

        # We might prefer AcceptRole for the 'select button', but that caused
        # default button grief.  In particular, it made 'Select' a default
        # button, but that caused 'enter' on the search edit to also select in
        # addition to refreshing the search.
        select = buttonbox.addButton("Selec&t", QtGui.QDialogButtonBox.ActionRole)
        buttonbox.addButton(QtGui.QDialogButtonBox.Cancel)

        select.clicked.connect(self.selectCurrent)
        buttonbox.rejected.connect(self.close)
        self.table.activated.connect(self.select)

        self.table.model().reset_content_from_session()
        
        self.geo = WindowGeometry(self, position=False)

    def select(self, index):
        self.selected_id = self.table.model().objectConverter(index.internalPointer())
        self.close()

    def selectedItem(self, class_, session):
        if hasattr(self,"selected_id"):
            return session.query(class_).filter(self.entity.key_column==self.selected_id).one()

    def selectCurrent(self):
        self.select(self.table.selectedIndexes()[0])

class PBMdiTableView(PBSearchableListDialog):
    """
    This class is now deprecated in favor of :class:`qtalchemy.ext.QueryDataView`.
    """
    def __init__(self, ssrc, entityCls):
        extensionId = "%s/MDISearch" % (entityCls.__name__, )

        PBSearchableListDialog.__init__(self, extensionId=extensionId)

        self.ssrc = ssrc
        self.entity = entityCls(ssrc, self)
        self.base_query, converter = self.entity.list_query_converter()
        self.table.setModel(QueryTableModel(self.base_query, ssrc=ssrc,objectConverter=converter),
                            extensionId=suffixExtId(self, "Table"))
        self.bindings = self.entity.itemCommands.withView(self.table, bindDefault=True)

        self.table.model().reset_content_from_session()

    def keyPressEvent(self, e):
        if e.key() == QtCore.Qt.Key_Escape:
            e.ignore()
        else:
            PBSearchableListDialog.keyPressEvent(self, e)

def colAttr(col):
    return col.property.columns[0].name

class PBTableTab(QtGui.QWidget):
    """
    This class is now deprecated in favor of :class:`qtalchemy.ext.QueryDataView`.
    """
    def __init__(self, parent, ssrc, entityCls, filters=None, query=None, extensionId=None, objectConverter=None):
        QtGui.QWidget.__init__(self, parent)

        self.ssrc = ssrc
        self.entity = entityCls(ssrc, parent)
        if query is None:
            self.query, self.objectConverter = self.entity.list_query_converter()
        else:
            self.query = query
            self.objectConverter = objectConverter
        self.filters = filters

        if extensionId is None:
            extensionId = suffixExtId(parent, entityCls.__name__)
        if extensionId is not None:
            self.setProperty("ExtensionId", extensionId)

        layout = QtGui.QVBoxLayout()
        self.setLayout(layout)

        self.toolbar = LayoutWidget(layout, QtGui.QToolBar(self))
        # build the table and attach a model
        self.table = LayoutWidget(layout,TableView(extensionId=suffixExtId(self, "Table")))
        self.table.setModel(QueryTableModel(self.query, ssrc=ssrc, objectConverter=self.objectConverter), 
                            extensionId=suffixExtId(self, "Table"))

        self.bindings = self.entity.itemCommands.withView(self.table, bindDefault=True)
        self.bindings.fillToolbar(self.toolbar)

        dataContextNeeded = False
        if self.filters is not None:
            for col, value in self.filters:
                if callable(value):
                    dataContextNeeded = True

        if not dataContextNeeded:
            self.refresh()

    def refresh(self, dataContext=None):
        q=self.query
        if self.filters is not None:
            for col, value in self.filters:
                if callable(value):
                    value = value(dataContext)
                q = q.filter(col==value)
                self.entity.info[colAttr(col)] = value
        self.table.model().query = q

        self.table.model().reset_content_from_session()



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2011, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################

from qtalchemy.widgets import PBKeyEdit
from user_attr import UserAttr
from PyQt4 import QtCore, QtGui
from sqlalchemy.orm import mapper, create_session, relation, object_session, Query
from PyQtModels import QueryTableModel, attrReadonly
from input_yoke import InputYoke

class RowWidgetEventProxy(object):
    """
    This proxy manages call-backs from bound widgets and forwards them to a 
    virtual method in the column definition.
    """
    def __init__(self, widget, row, userattr):
        self.widget = widget
        self.row = row
        self.userattr = userattr

    def search(self):
        self.userattr.search(self.widget, self.row)

class ForeignKeyReferral(UserAttr):
    """
    A ForeignKeyReferralEx UserAttr provides a display for a user-centric key 
    matching back to a database foreign key which is not expected to be 
    user-friendly.  This also wraps in the ability for search method which is 
    expected to display a dialog with a list of entities which can be chosen 
    in this attribute.

    :param atype:  python type of displayed key
    :param label:  widget label for UI
    :param backref:  python attribute name in the ModelObject holding the referred key object
    :param class_:  python class of referred key object
    :param userkey:  python attribute name of class_ to display or callable taking parameter of type class_
    :param entity:  The domain entity class this foreign key represents a member of.
    """
    def __init__(self, atype, label, backref, class_, userkey, entity=None, filter_query=None, readonly=False):
        self.backref = backref
        self.class_ = class_
        self.userkey = userkey
        self.filter_query = filter_query
        self.entityCls = entity
        UserAttr.__init__(self,atype,label,readonly=readonly)
        self.proxies = {}

    def fget(self,row):
        target = getattr(row,self.backref)
        if target is None:
            return None
        if callable(self.userkey):
            return self.userkey(target)
        else:
            return getattr(target, self.userkey)

    def fset(self,row,value):
        if value is None:
            # for null assignments, we don't need to look at the database
            setattr(row,self.backref,None)
            return
        s = self.session(row)
        try:
            # we use ilike to complement the QtCore.Qt.CaseInsensitive in the QCompleter
            setattr(row,self.backref,s.query(self.class_).filter(getattr(self.class_,self.userkey).ilike(value)).one())
        except Exception as e:
            print e
            setattr(row,self.backref,None)

    def getSelection(self):
        q=Query((getattr(self.class_,self.userkey).label("name"),))
        if self.filter_query:
            q = self.filter_query(q)
        return q

    def yoke_specifier(self):
        if self.entityCls is None:
            return "foreign_key_combo"
        else:
            return "foreign_key"

    def WidgetFactory(self,parent, index):
        # Note that it's not entirely clear to me what criteria is appropriate for selecting a PBKeyEdit or a QComboBox.
        if self.entityCls is not None:
            w = PBKeyEdit(parent)
            w.model_index = index
            # The PBKeyEdit has a 'search button' which offers a more comprensive notion of a list of options.
            # The attached entityCls is expected to know what data to display in the search list.
        else:
            # Some days we think this should be a combo-box.
            w = QtGui.QComboBox(parent)
            w.setEditable(True)
        return w

    def WidgetBind(self,widget,row):
        Session = self.session_maker(row)
        # TODO: this is a very bad performance problem on large data-sets.
        # I suspect we need a highly customized completer so as to only load the parts with this prefix.
        col_model = QueryTableModel(self.getSelection(),ssrc=Session)
        col_model.reset_content_from_session()

        # note that a PBKeyEdit is a QLineEdit
        if isinstance(widget, PBKeyEdit):
            # A widget can never be associated with more than one proxy.
            # I believe this is a natural sort of limitation as any specific bound 'edit-type' widget can only be bound to one row at a time.
            p = RowWidgetEventProxy(widget, row, self)
            self.proxies[id(widget)] = p # need to hold the proxies in memory
            widget.buttonPressed.connect(p.search)

        # I handle both QLineEdit and QComboBox here.
        if isinstance(widget,QtGui.QLineEdit):
            col_completer = QtGui.QCompleter()

            # The default QtGui.QCompleter.PopupCompletion is perhaps better
            # QtGui.QCompleter.InlineCompletion mode made the following bug:
            #  1)  tab through a dialog to PBKeyEdit on a table
            #  2)  observe that the text is fully selected in the PBKeyEdit
            #  3)  press enter
            #  4)  text updated in the model to the first item in col_model
            #  We presume that enter triggered the completion with the text before the input block (i.e. an empty string)
            #col_completer.setCompletionMode(QtGui.QCompleter.InlineCompletion)
            col_completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
            widget.setCompleter(col_completer)
            #widget.setValidator(ForeignKeyValidator(fk=col_type,session=session))
            col_completer.setModel(col_model)
        elif isinstance(widget,QtGui.QComboBox):
            widget.setModel(col_model)

    def search(self, widget, row):
        Session = self.session_maker(row)

        index = None
        parent = None
        try:
            index = widget.model_index
            parent = widget.parent()
        except:
            pass

        import PBSearchDialog
        srch = PBSearchDialog.PBSearchDialog(Session, self.entityCls, widget.parent())
        srch.setWindowModality(QtCore.Qt.ApplicationModal)
        srch.show()
        srch.exec_()
        result = srch.selectedItem(self.class_,self.session(row))
        if result is not None:
            if index is not None:
                index.model().ensureNoFlipper(index)

            # we set the data internally
            setattr(row, self.backref, result)
            #setattr(row, self.backref, srch.selected_id)
            #print "Selected item:  {0}\n".format(self.fget(row))

            if index is None:
                # We are not a widget associated with an abstract item view and so we suppose we are a visual element on a "dialog box lifespan".
                # Therefore the widget needs to be updated with the visual identifier of this foreign key.
                widget.setText(self.fget(row))
            else:
                #index.model().setData(index, self.fget(row))
                parent.setFocus()

class ForeignKeyEditYoke(InputYoke):
    def __init__(self,mapper,attr):
        InputYoke.__init__(self,mapper)
        self.attr = attr
        mapper.reverse_yoke(attr,self)

    def Factory(self):
        self.widget = PBKeyEdit()
        self._baseAdoptWidget(self.widget)
        self.widget.editingFinished.connect(self.Save)
        self.widget.buttonPressed.connect(self.Search)
        if attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)
        return self.widget

    def AdoptWidget(self, widget):
        self.widget = widget
        self._baseAdoptWidget(self.widget)
        self.widget.editingFinished.connect(self.Save)
        self.widget.buttonPressed.connect(self.Search)
        if attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)

    def Search(self):
        user_attr = getattr(self.mapper.cls,self.attr)
        row = self.mapper.obj
        Session = user_attr.session_maker(row)

        index = None
        parent = None
        try:
            index = self.widget.model_index
            parent = self.widget.parent()
        except:
            pass

        import PBSearchDialog
        srch = PBSearchDialog.PBSearchDialog(Session, user_attr.entityCls, self.widget.parent())
        srch.setWindowModality(QtCore.Qt.ApplicationModal)
        srch.show()
        srch.exec_()
        result = srch.selectedItem(user_attr.class_,user_attr.session(row))
        if result is not None:
            setattr(row, user_attr.backref, result)
            #setattr(row, self.backref, srch.selected_id)
            #print "Selected item:  {0}\n".format(self.fget(row))

            if index is None:
                # We are not a widget associated with an abstract item view and so we suppose we are a visual element on a "dialog box lifespan".
                # Therefore the widget needs to be updated with the visual identifier of this foreign key.
                self.widget.setText(user_attr.fget(row))
            else:
                #index.model().setData(index, user_attr.fget(row))
                parent.setFocus()

    def Bind(self):
        user_attr = getattr(self.mapper.cls,self.attr)
        Session = user_attr.session_maker(self.mapper.obj)

        # TODO: this is a very bad performance problem on large data-sets.
        # I suspect we need a highly customized completer so as to only load the parts with this prefix.
        col_model = QueryTableModel(user_attr.getSelection(),ssrc=Session)
        col_model.reset_content_from_session()

        col_completer = QtGui.QCompleter()

        col_completer.setCompletionMode(QtGui.QCompleter.InlineCompletion)
        col_completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)
        self.widget.setCompleter(col_completer)
        #widget.setValidator(ForeignKeyValidator(fk=col_type,session=session))
        col_completer.setModel(col_model)

        if isinstance(self.widget, QtGui.QLineEdit):
            self.widget.setText(self.mapper.getObjectAttr(self.attr))
        else:
            self.widget.setEditText(self.mapper.getObjectAttr(self.attr))

    def Save(self):
        if attrReadonly(self.mapper.cls, self.attr):
            return
        if isinstance(self.widget, QtGui.QLineEdit):
            self.mapper.setObjectAttr(self.attr,self.widget.text())
        else:
            self.mapper.setObjectAttr(self.attr,self.widget.currentText())

class ForeignKeyComboYoke(InputYoke):
    def __init__(self,mapper,attr):
        InputYoke.__init__(self,mapper)
        self.attr = attr
        mapper.reverse_yoke(attr,self)

    def Factory(self):
        self.widget = QtGui.QComboBox()
        self._baseAdoptWidget(self.widget)
        self.widget.setEditable(True)
        #self.widget.editingFinished.connect(self.Save)
        return self.widget

    def AdoptWidget(self, widget):
        self.widget = widget
        self._baseAdoptWidget(self.widget)
        self.widget.setEditable(True)
        if hasattr(self.widget, "setReadOnly") and attrReadonly(self.mapper.cls, self.attr):
            self.widget.setReadOnly(True)
        #self.widget.editingFinished.connect(self.Save)
        return self.widget

    def Bind(self):
        user_attr = getattr(self.mapper.cls,self.attr)
        Session = user_attr.session_maker(self.mapper.obj)

        col_model = QueryTableModel(user_attr.getSelection(),ssrc=Session)
        col_model.reset_content_from_session()

        self.widget.setModel(col_model)
        self.widget.setEditText(self.mapper.getObjectAttr(self.attr))

    def Save(self):
        if attrReadonly(self.mapper.cls, self.attr):
            return
        self.mapper.setObjectAttr(self.attr,self.widget.currentText())

###
# I really want this ForeignKeyValidator to give aggressive foreign key validation, but QValidator evidently confuses me.
###
class ForeignKeyValidator(QtGui.QValidator):
    def __init__(self,parent=None,fk=None,session=None):
        QtGui.QValidator.__init__(self,parent)
        self.fkr = fk
        self.session = session

    def fixup(self,input):
        print "f"
        like_me = self.session.query(self.fkr.header_name).filter(self.fkr.header_name.ilike(str(input))).all()
        if len(like_me) == 1:
            print QtCore.QString(like_me[0].vid)
            input.replace(0,6,QtCore.QString(like_me[0].vid))
            print input, type(input)
        #return str(input)

    def validate(self,input,pos):
        print "v"
        try:
            like_me = self.session.query(self.fkr.header_name).filter(self.fkr.header_name.ilike(str(input))).one()
            print "v2"
            input.replace(0,6,QtCore.QString(like_me[0].vid))
            return QtGui.QValidator.Acceptable,pos
        except:
            return QtGui.QValidator.Intermediate,pos



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
The search dialog is a very high level GUI element of the qtalchemy library.  It
utilizes the DomainEntity to provide descriptive details of the objects to query
in these search dialogs. 
"""

from qtalchemy import *
from sqlalchemy.orm import mapper, create_session, relation, object_session
import sqlalchemy.sql.expression as expr
from PyQt4 import QtGui, QtCore, QtDeclarative

class QuickList(QtGui.QWidget):
    """
    This base class gathers a search edit and table together in a single dialog.
    """
    def __init__(self, qml, ssrc, entityCls, parent=None, extensionId=None):
        QtGui.QWidget.__init__(self, parent, QtCore.Qt.Dialog)

        self.setProperty("ExtensionId", extensionId)

        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        searchLayout = QtGui.QHBoxLayout()
        vbox.addLayout(searchLayout)
        searchLabel = QtGui.QLabel("&Search: ")
        searchLayout.addWidget(searchLabel)
        self.searchEdit = QtGui.QLineEdit()
        self.searchEdit.returnPressed.connect(self.resetSearch)
        searchLabel.setBuddy(self.searchEdit)
        searchLayout.addWidget(self.searchEdit)
        self.searchGo = QtGui.QPushButton("&Refresh")
        searchLayout.addWidget(self.searchGo)
        self.searchGo.clicked.connect(self.resetSearch)

        self.ssrc = ssrc
        self.entity = entityCls(ssrc, self)
        self.base_query, converter = self.entity.list_query_converter()

        self.model = QueryTableModel(self.base_query, ssrc=ssrc, objectConverter=converter)
        
        self.view = LayoutWidget(vbox,QtDeclarative.QDeclarativeView())
        context = self.view.rootContext()
        context.setContextProperty("myModel", self.model)
        self.view.setSource(QtCore.QUrl.fromLocalFile(qml))
        
        self.setFocusProxy(self.searchEdit)

    def resetSearch(self):
        if self.searchEdit.text() != "":
            like_str = "%%%s%%"%(self.searchEdit.text(),)
            filter_cols = [c.ilike(like_str) for c in self.entity.list_search_columns]

            self.model.query = self.base_query.filter(expr.or_(*tuple(filter_cols)))
        else:
            self.model.query = self.base_query

        self.model.reset_content_from_session()

    def keyPressEvent(self, e):
        if e.key() == QtCore.Qt.Key_Down:
            self.view.setFocus(QtCore.Qt.TabFocusReason)
            e.accept()
        else:
            QtGui.QDialog.keyPressEvent(self, e)



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################

from qtalchemy import *
from qtalchemy.widgets import TableView
from qtalchemy.dialogs import *
from PyQt4 import QtCore, QtGui
from sqlalchemy.orm.attributes import InstrumentedAttribute
import os.path
import decimal
import datetime
import fuzzyparsers

class ImportColumn(ModelObject):
    def __init__(self, identifier, label):
        self.identifier = identifier
        self.label = label
    
    identifier = UserAttr(str, "Identifier")
    label = UserAttr(str, "Label")

class ImportIntroPage(QtGui.QWizardPage):
    """
    This wizard page takes a class and displays the elements of that class for 
    import.
    """
    def __init__(self, parent=None):
        QtGui.QWizardPage.__init__(self, parent)
        self.setTitle("Data Import")
        self.setSubTitle("\
The list of available fields in the import class are shown below.  Columns in \
import files are first matched with exact matches in the identifier list and then \
matched against prefixes of the labels.")

        main = QtGui.QVBoxLayout(self)
        self.table = LayoutWidget(main, TableView())
        self.model = ClassTableModel(ImportColumn, ("identifier", "label"))
        self.table.setModel(self.model)

    def loadImportColumns(self, cls, attrList=None):
        if attrList is None:
            attrList = []
            for attr in dir(cls):
                col = getattr(cls, attr)
                try:
                    x = attrLabel(cls, attr)
                    if x not in [None, ""] and not attrReadonly(cls, attr):
                        attrList.append(attr)
                except Exception as e:
                    # if we can't find a nice label, we skip it
                    pass

        items = [ImportColumn(attr, attrLabel(cls, attr)) for attr in attrList]
        items.sort(key=lambda x: x.label)
        self.model.reset_content_from_list(items)

class ImportDataSource(QtGui.QWizardPage):
    def __init__(self, parent=None):
        QtGui.QWizardPage.__init__(self, parent)
        self.setTitle("Data Import")
        self.setSubTitle("Select a file from which to import the data.")

        main = QtGui.QVBoxLayout(self)
        fileSelection = LayoutLayout(main, QtGui.QHBoxLayout())
        self.label = LayoutWidget(fileSelection, QtGui.QLabel("&Data File:"))
        self.fileEdit = LayoutWidget(fileSelection, QtGui.QLineEdit())
        self.fileEdit.textChanged.connect(self.completeChanged.emit)
        self.label.setBuddy(self.fileEdit)
        LayoutWidget(fileSelection, QtGui.QPushButton("Bro&wse...")).clicked.connect(self.import_browse)

    def import_browse(self):
        fileName = QtGui.QFileDialog.getOpenFileName(self,"Data for Import",filter="Comma Separated Values (*.csv);;All Files (*.*)")
        if fileName != "":
            self.fileEdit.setText(fileName)
            self.completeChanged.emit()

    def isComplete(self):
        return os.path.isfile(self.fileEdit.text())

class ImportDataPreview(QtGui.QWizardPage):
    def __init__(self, parent=None):
        QtGui.QWizardPage.__init__(self, parent)
        self.setTitle("Data Import")
        self.setSubTitle("Preview the imported data.")

        main = QtGui.QVBoxLayout(self)
        self.table = LayoutWidget(main, TableView())
        
        self.session = None

    def loadData(self, Session, cls, csvFile):
        import csv
        
        if self.session is not None:
            self.session.close()
            self.session = None

        self.session = Session()
        importObjects = []
        
        errors = []
        cols = None
        index = 0
        for row in csv.reader(open(csvFile, "r")):
            if cols is None:
                cols = []
                for item in row:
                    cols.append(item.strip())
                self.model = ClassTableModel(cls, cols)
                self.table.setModel(self.model)
            else:
                try:
                    index += 1
                    c = cls()
                    self.session.add(c)
                    importObjects.append(c)
                    for i in range(len(cols)):
                        type_ = ClassAttributeType(getattr(cls, cols[i]))
                        v = row[i]
                        if type_ in (decimal.Decimal, int, float) and v=="":
                            v = "0"
                        if type_ in (datetime.date, ):
                            v = fuzzyparsers.parse_date(v)
                        if not isinstance(v, type_):
                            v = type_(v)
                        #print row[i], type_, type_(row[i])
                        setattr(c, cols[i], v)
                except Exception as e:
                    errors.append("Row {0}:  {1}".format(index, str(e)))

        if len(errors) > 0 and len(errors) <= 5:
            errMsg = """\
There were {0} errors importing the data.

<ul>
{1}
</ul>""".format(len(errors), '\n'.join(["<li>{0}</li>".format(e) for e in errors]))
            appMessage(self, Message.Ok, errMsg)
        elif len(errors) > 5:
            errMsg = """\
There were {0} errors importing the data.  The first 5 are shown below.

<ul>
{1}
</ul>""".format(len(errors), '\n'.join(["<li>{0}</li>".format(e) for e in errors[:5]]))
            appMessage(self, Message.Ok, errMsg)

        self.model.reset_content_from_list(importObjects)

    def finishImport(self):
        messaged_commit(self.session, self)
        self.session.close()
        self.session = None
        

class ImportWizard(QtGui.QWizard):
    """
    Returns a QWizard to take rows given by the user and create an instance 
    of each cls from each row.  The rows may come from a csv file or via raw 
    entry in a table.

    Wizard Structure:

    - Intro page listing accepted elements of the class cls
    - Page asking for an input csv file or allowing raw entry
    - Page with list of data or allowing entry
    
    >>> from sqlalchemy import Table, Column, String, Integer, MetaData, create_engine
    >>> from sqlalchemy.ext.declarative import declarative_base
    >>> 
    >>> from qtalchemy import *
    >>> from qtalchemy.ext.dataimport import *
    >>> from PyQt4 import QtCore, QtGui
    >>> 
    >>> metadata = MetaData()
    >>> Base = declarative_base(metadata=metadata, cls=ModelObject)
    >>> 
    >>> class People(Base):
    ...     __table__ = Table('people', metadata,
    ...                     Column('id', Integer, primary_key=True),
    ...                     Column('f_name', String(50), info={"label": "First Name"}),
    ...                     Column('l_name', String(50), info={"label": "Last Name"}),
    ...                     Column('title', String(10), info={"label": "Title"}))
    ... 
    >>> engine = create_engine("sqlite://")
    >>> metadata.bind = engine
    >>> metadata.create_all()
    >>> Session = PBSessionMaker(bind=engine)
    >>> 
    >>> app = qtapp()
    >>> wiz = ImportWizard(Session, People)
    >>> wiz.show()  #doctest: +SKIP
    >>> wiz.exec_()  #doctest: +SKIP
    """
    def __init__(self, Session, cls, parent=None):
        QtGui.QWizard.__init__(self, parent)

        self.Session = Session
        self.cls = cls

        self.setWindowTitle("Import Wizard")

        self.intro = ImportIntroPage()
        self.intro.loadImportColumns(cls)

        self.dataSource = ImportDataSource()

        self.dataPreview = ImportDataPreview()

        self.introId = self.addPage(self.intro)
        self.sourceId = self.addPage(self.dataSource)
        self.previewId = self.addPage(self.dataPreview)

        self.currentIdChanged.connect(self.pageFlip)
        self.accepted.connect(self.dataPreview.finishImport)

    def pageFlip(self, newid):
        if newid == self.previewId:
            self.dataPreview.loadData(self.Session, self.cls, self.dataSource.fileEdit.text())



# -*- coding: utf-8 -*-
"""
We create models, attach to data and see if the events all propogate.
"""

from qtalchemy import *
from PyQt4 import QtCore, QtGui

class ItemEntity(object):
    contextMenu = CommandMenu("_context_menu")
    
    @contextMenu.command("View Info", sort=200)
    def view_info(self, id):
        pass

    @contextMenu.command("Print Ticket")
    def print_ticket(self, id):
        pass

    @contextMenu.command("New Invoice", sort=400)
    def new_invoice(self, id):
        pass

def items_there_test():
    x = ItemEntity()
    assert len(x.contextMenu), "3 commands were added"
    assert x.contextMenu[0].descr.lower().startswith("view"), "view was explicitly put first"
    assert x.contextMenu[1].descr.lower().startswith("new"), "new was explicitly put next"



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################

# Qt resource file
import icons_rc

from button_edit import PBButtonEdit
from date_edit import PBDateEdit
from key_edit import PBKeyEdit

from table import TableView
from tree import TreeView

from street_address_widget import \
    StreetAddress, StreetAddressYoke, StreetAddressEdit, parse_address, concat_address



# -*- coding: utf-8 -*-
from qtalchemy import ModelObject, UserAttr
from PyQt4 import QtCore, QtGui

class AuthSettings(ModelObject):
    """
    :class:`AuthSettings` contains the attributes for the model to 
    attach to the authentication dialogs.
    """
    classEvents = ModelObject.Events()

    remember_server = UserAttr(bool, "Remember Data Source")
    remember_user = UserAttr(bool, "Remember User")
    
    server = UserAttr(str, "Server")
    database = UserAttr(str, "Database")
    user = UserAttr(str, "User")
    password = UserAttr(str, "Password")
    server_string = UserAttr(str, "Server Configuration", readonly=True)
    
    def __init__(self):
        self.remember_server = True
        self.remember_user = True
        self.server_type = "postgresql"
        self.server = None
        self.database = None
        self.user = None
        self.password = None

    @classEvents.add("set", "server")
    @classEvents.add("set", "database")
    def server_string_config_change(self, attr, oldvalue):
        self.server_string = "%s://%s/%s" % (self.server_type, self.server, self.database)

    def loadSettings(self):
        settings = QtCore.QSettings()

        settings.beginGroup("Data Settings (Authentication)")
        self.remember_server = settings.value("remember_server",self.remember_server)
        self.remember_user = settings.value("remember_user",self.remember_user)
        if self.remember_server:
            self.server_type = settings.value("server_type",self.server_type)
            self.server = settings.value("server",self.server)
            self.database = settings.value("database",self.database)
        if self.remember_user:
            self.user = settings.value("user",self.user)
        settings.endGroup()

    def saveSettings(self):
        settings = QtCore.QSettings()

        settings.beginGroup("Data Settings (Authentication)")
        settings.setValue("remember_server",self.remember_server)
        settings.setValue("remember_user",self.remember_user)
        settings.setValue("server_type",self.server_type if self.remember_server else "")
        settings.setValue("server",self.server if self.remember_server else "")
        settings.setValue("database",self.database if self.remember_server else "")
        settings.setValue("user",self.user if self.remember_user else "")
        settings.endGroup()



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
The qtalchemy library builds a database application framework combining 
sqlalchemy and PyQt/PySide.
"""

# TODO:  Review for best use of our global namespace to highlight the correct 
# way to do things.

from PyQtHelpers import LayoutWidget, LayoutLayout, ButtonBoxButton, FormRow, \
            WindowGeometry, OnlineHelp, \
            writeTableColumnGeo, readTableColumnGeo, suffixExtId, \
            fromQType, toQType, qtapp, message_excepthook, \
            Signal, Slot, Property

from commands import DomainEntity, Command, CommandMenu, CommandEvent, \
        BoundCommandMenu, BoundCommandMenuItem

from user_attr import UserAttr, Nullable, AttrNumeric, UseCachedValue
from sqlalchemy_helper import ModelObject, ModelSession, PBSessionMaker, UUID, \
        ValidationError, user_message, Message, instanceEvent, sessionExtension

from input_yoke import \
    addGlobalYoke, \
    InputYoke, \
    LineYoke, \
    SelectionYoke, \
    TextYoke, \
    IntegerYoke, \
    DateYoke, \
    FloatingPointYoke

# model elements
from PyQtModels import ClassTableModel, QueryTableModel, QueryClassTableModel, \
        ModelColumn, PBTableModel, AlchemyModelDelegate, modelMimeRectangle, \
        MapperMixin, WidgetAttributeMapper, ObjectRepresent, \
        attrLabel, ClassAttributeLabel, attrType, ClassAttributeType, attrReadonly

from foreign_key import ForeignKeyReferral, ForeignKeyEditYoke, ForeignKeyComboYoke

# TODO:  Some more of these should be moved to the widgets sub-module
# gui elements
from PBSearchDialog import PBSearchDialog, PBTableTab, PBMdiTableView

import icons_rc



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
The simplest use of the tools in the dataimport module is to create a wizard which 
uses a session-maker object and a class and imports data from a file or 
by raw tabular entry.
"""

from wizard import ImportWizard, \
    ImportDataSource, \
    ImportDataPreview, \
    ImportIntroPage



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################

from PyQt4 import QtGui,QtCore
import re
from qtalchemy import *

from button_edit import *  # for Signal and Property

def is_country(s):
    """
    Return true if the string s looks like a country name.

    We only check that the string doesn't have numbers in it (very lame for the moment).
    """
    return re.search('[^ a-zA-Z]',s) is None

def city_state(s):
    """
    >>> city_state('bethlehem pa')
    ('Bethlehem', 'PA')
    >>> city_state('seattle, washington')
    ('Seattle', 'Washington')
    >>> city_state('New york, new york')
    ('New York', 'New York')
    """
    s = s.strip()
    if s.find(',')>-1:
        city,state = s.rsplit(',',2)
    elif s.find(' ')>-1:
        city,state = s.rsplit(' ',2)
    else:
        city,state = s,None
    city = city.strip()
    state = state.strip()
    if state is not None and len(state)==2:
        state = state.upper()
    else:
        state = state.title()
    city = city.title()
    return city,state

def parse_address(addr):
    """
    Parses a single string into street, city, state, zip and country.  
    Semi-colons, new lines, and carriage returns are all recognized as logical 
    line breaks.  Parsing proceeds roughly from right to left recognizing the 
    country first (if there), zip/postal code, state/province, city, and finally 
    street address components.  Two street addresses are recognized for PO Box 
    lines or other street details.

    >>> parse_address('5522 Penelope Parkway; dayton, me 12345')
    ('5522 Penelope Parkway', None, 'Dayton', 'ME', '12345', 'USA')
    >>> parse_address('PO box 14; 103 Kaiser Lane; Marksville, AK 98765')
    ('PO Box 14', '103 Kaiser Lane', 'Marksville', 'AK', '98765', 'USA')
    >>> parse_address('531 fiddley blvd; Marksville, AB K1A 0B1')
    ('531 Fiddley Blvd', None, 'Marksville', 'AB', 'K1A 0B1', 'Canada')
    """
    lines = [l for l in re.split('(;|\n|\r)',addr) if l not in list(';\n\r')]
    address1=address2=city=state=zip=country=None
    items = ("street","zip","country")
    while len(lines)>0 and len(items)>0:
        if items[-1] == "country":
            if is_country(lines[-1]):
                country = lines[-1]
                lines = lines[:-1]
            items = items[:-1]
            continue
        elif items[-1] == "zip":
            if re.search('[0-9]{5}$',lines[-1].strip()):
                # USA zip
                if country is None:
                    country="USA"
                zip = lines[-1].strip()[-5:]
                city, state = city_state(lines[-1].strip()[:-5])
                lines = lines[:-1]
                items = items[:-1]
            elif re.search('[0-9]{5}-[0-9]{4}$',lines[-1].strip()):
                # USA zip
                if country is None:
                    country="USA"
                zip = lines[-1].strip()[-10:]
                city, state = city_state(lines[-1].strip()[:-10])
                lines = lines[:-1]
                items = items[:-1]
            elif re.search('[0-9a-zA-Z]{3} [0-9a-zA-Z]{3}$',lines[-1].strip()):
                # Canada zip
                if country is None:
                    country="Canada"
                zip = lines[-1].strip()[-7:]
                city, state = city_state(lines[-1].strip()[:-7])
                lines = lines[:-1]
                items = items[:-1]
            else:
                items = items[:-1]
            continue
        else:
            if len(lines)==1:
                address1 = lines[0].strip().title()
            elif len(lines)==2:
                address1,address2 = lines[0].strip().title(),lines[1].strip().title()
            else:
                raise Exception("address can't be parsed")
            lines=[]
    if address1 is not None and address1.startswith('Po Box'):
        address1 = 'PO Box' + address1[6:]
    if address2 is not None and address2.startswith('Po Box'):
        address2 = 'PO Box' + address2[6:]
    return (address1,address2,city,state,zip,country)

def concat_address(addr1, addr2, city, state, zip, country, linebreak='\n'):
    '''
    This reverses the parsing action of :func:parse_address joining the 
    components of an address into a single string.
    
    >>> concat_address('230 E. Main Street', None, 'Bethlehem', 'OH', '18030', '', linebreak='; ')
    '230 E. Main Street; Bethlehem OH 18030'
    '''
    lines = []
    if addr1 not in (None,''):
        lines.append(addr1)
    if addr2 not in (None,''):
        lines.append(addr2)
    lines.append(' '.join([v for v in (city,state,zip) if v not in (None, '')]))
    if country not in (None,''):
        lines.append(country)
    return linebreak.join(lines)

class StreetAddressEdit(QtGui.QStackedWidget):
    addressParsed = Signal(name='addressParsed')

    def __init__(self,parent=None):
        QtGui.QStackedWidget.__init__(self,parent)

        self.concat_page = QtGui.QWidget()
        self.concat_page.setObjectName("concat_page")
        vbox = QtGui.QVBoxLayout(self.concat_page)
        vbox.setContentsMargins(1,1,1,1)
        self.unparsedAddress_edit = LayoutWidget(vbox,QtGui.QTextEdit(self.concat_page))
        self.unparsedAddress_edit.setObjectName("unparsedAddress_edit")
        hbox = LayoutLayout(vbox,QtGui.QHBoxLayout())
        self.btnParse = LayoutWidget(hbox,QtGui.QPushButton("Parse (F4)",self.concat_page))
        size = self.btnParse.minimumSize()
        size.setWidth(140)
        self.btnParse.setMinimumSize(size)
        hbox.addItem(QtGui.QSpacerItem(375, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum))
        self.addWidget(self.concat_page)

        self.parsed_page = QtGui.QWidget()
        grid = QtGui.QGridLayout(self.parsed_page)
        grid.setContentsMargins(0,0,0,0)
        grid.setObjectName("gridLayout")
        self.address1_edit = QtGui.QLineEdit(self.parsed_page)
        self.address1_edit.setObjectName("address1_edit")
        grid.addWidget(self.address1_edit, 0, 0, 1, 3)
        self.address2_edit = QtGui.QLineEdit(self.parsed_page)
        self.address2_edit.setObjectName("address2_edit")
        grid.addWidget(self.address2_edit, 1, 0, 1, 3)
        self.city_edit = QtGui.QLineEdit(self.parsed_page)
        self.city_edit.setObjectName("city_edit")
        grid.addWidget(self.city_edit, 3, 0, 1, 1)
        self.state_edit = QtGui.QLineEdit(self.parsed_page)
        self.state_edit.setMaximumSize(QtCore.QSize(40, 16777215))
        self.state_edit.setObjectName("state_edit")
        grid.addWidget(self.state_edit, 3, 1, 1, 1)
        self.zip_edit = QtGui.QLineEdit(self.parsed_page)
        self.zip_edit.setMaximumSize(QtCore.QSize(100, 16777215))
        self.zip_edit.setObjectName("zip_edit")
        grid.addWidget(self.zip_edit, 3, 2, 1, 1)
        self.country_edit = QtGui.QLineEdit(self.parsed_page)
        self.country_edit.setObjectName("country_edit")
        grid.addWidget(self.country_edit, 4, 1, 1, 2)
        self.btnConcatenate = QtGui.QPushButton("Concatenate (F4)",self.parsed_page)
        self.btnConcatenate.setObjectName("btnConcatenate")
        size = self.btnConcatenate.minimumSize()
        size.setWidth(140)
        self.btnConcatenate.setMinimumSize(size)
        grid.addWidget(self.btnConcatenate, 4, 0, 1, 1, QtCore.Qt.AlignLeft)
        self.addWidget(self.parsed_page)

        self.parseAct = QtGui.QAction("parse/concat",self)
        self.parseAct.setShortcut(QtCore.Qt.Key_F4)
        self.parseAct.triggered.connect(self.parse_or_concat)
        self.addAction(self.parseAct)

        self.btnParse.clicked.connect(self.toParsed)
        self.btnConcatenate.clicked.connect(self.toConcat)

    def parse_or_concat(self):
        if self.currentIndex() == 0:
            self.toParsed()
        else:
            self.toConcat()

    def toParsed(self):
        try:
            address1,address2,city,state,zip,country = parse_address(self.unparsedAddress_edit.toPlainText())
            self.address1_edit.setText(address1)
            self.address2_edit.setText(address2)
            self.city_edit.setText(city)
            self.state_edit.setText(state)
            self.zip_edit.setText(zip)
            self.country_edit.setText(country)
            self.setCurrentIndex(1)
            self.addressParsed.emit()
        except Exception, e:
            QtGui.QMessageBox.information(self.parent(),QtGui.QApplication.applicationName(),str(e))

    def toConcat(self):
        addr = concat_address(
                    self.address1_edit.text(),
                    self.address2_edit.text(),
                    self.city_edit.text(),
                    self.state_edit.text(),
                    self.zip_edit.text(),
                    self.country_edit.text())
        self.unparsedAddress_edit.setPlainText(addr)
        self.setCurrentIndex(0)

    def implantProperty(self,prop,v):
        if self.currentIndex() == 0:
            self.toParsed()
        #print "Set {0}={1}".format(prop,v)
        getattr(self, prop+"_edit").setText(v)

    def extractProperty(self,prop):
        if self.currentIndex() == 0:
            t = parse_address(self.unparsedAddress_edit.toPlainText())
            index = 'address1,address2,city,state,zip,country'.split(',').index(prop)
            #print "Get (parsing) {0}={1}".format(prop,t[index])
            return t[index]
        else:
            #print "Get (easy) {0}={1}".format(prop,getattr(self, prop+"_edit").text())
            return getattr(self, prop+"_edit").text()

    address1 = Property("QString", 
            lambda self: self.extractProperty("address1"), 
            lambda self, v: self.implantProperty("address1", v))

    address2 = Property("QString", 
            lambda self: self.extractProperty("address2"), 
            lambda self, v: self.implantProperty("address2", v))

    city = Property("QString", 
            lambda self: self.extractProperty("city"), 
            lambda self, v: self.implantProperty("city", v))

    state = Property("QString", 
            lambda self: self.extractProperty("state"), 
            lambda self, v: self.implantProperty("state", v))

    zip = Property("QString", 
            lambda self: self.extractProperty("zip"), 
            lambda self, v: self.implantProperty("zip", v))

    country = Property("QString", 
            lambda self: self.extractProperty("country"), 
            lambda self, v: self.implantProperty("country", v))


class StreetAddressYoke(InputYoke):
    def __init__(self,mapper,attr):
        InputYoke.__init__(self,mapper)
        self.attr = attr
        user_attr = getattr(self.mapper.cls,self.attr)
        for x in user_attr.my_cols:
            mapper.reverse_yoke(x,self)
        self.trapRecurse = 0

    def Factory(self):
        self.widget = StreetAddressEdit()
        self.widget.addressParsed.connect(self.Save)
        self._baseAdoptWidget(self.widget)
        return self.widget

    def AdoptWidget(self, widget):
        raise NotImplementedError()

    def Bind(self):
        if self.trapRecurse > 0:
            return

        user_attr = getattr(self.mapper.cls,self.attr)
        
        parts = [self.mapper.getObjectAttr(a) for a in user_attr.my_cols]
        parts = ["" if p is None else p for p in parts]
        props = 'address1,address2,city,state,zip,country'.split(',')
        if parts == [""]*6:
            self.widget.setCurrentIndex(0)
        for i in range(6):
            self.widget.setProperty(props[i], parts[i])
        if parts != [""]*6:
            self.widget.setCurrentIndex(1)

    def Save(self):
        try:
            self.trapRecurse += 1
            user_attr = getattr(self.mapper.cls,self.attr)

            props = 'address1,address2,city,state,zip,country'.split(',')
            for i in range(6):
                self.mapper.setObjectAttr(user_attr.my_cols[i],self.widget.property(props[i]))
        finally:
            self.trapRecurse -= 1

class StreetAddress(UserAttr):
    """
    A StreetAddress UserAttr displays a street address in a stacked widget 
    allowing multi-line input which can be parsed into the constituent parts.

    TODO:  force the constituent parsing before allowing the bound attributes to passed off as valid.

    :param myCols:  A list of attribute names which store (in order):  
                address1, address2, city, state, zip, country
    """
    def __init__(self, label, myCols, whats_this=None):
        self.my_cols = myCols
        UserAttr.__init__(self, str, label, whats_this=whats_this)

    def fget(self,row):
        return concat_address(*map(lambda x: getattr(row,x),self.my_cols))

    def fset(self,row,value):
        raise NotImplementedError("StreetAddress types may not be assigned en-masse")

    def yoke_specifier(self):
        return 'address'



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
For the most part, we have functions here which do something for windows 
and something else for everything else.
"""

def is_windows():
    """
    returns True if running on windows
    """
    import sys
    return sys.platform in ('win32', 'cygwin')

def xdg_open(file):
    """
    Be a platform smart incarnation of xdg-open and open files in the correct 
    application.
    """
    import os
    if is_windows():
        try:
            # we try with win32api because that's cleaner (os.system flickers a command prompt)
            import win32api
            win32api.ShellExecute(0, None, file, None, None, 1)
        except ImportError as e:
            os.system('start {0}'.format(file))
    else:
        os.system('xdg-open {0}'.format(file.replace(';', r'\;').replace('&', r'\&')))



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
This module contains tools to enable cross platform support.

For the most part, we have functions here which do something for windows 
and something else for everything else.
"""

def guiexcepthook(type, value, tb):
    """
    Replace sys.excepthook with this function to display errors more gracefully 
    for an application which is not associated with a console.
    
    >>> import sys
    >>> from qtalchemy import xplatform
    >>> sys.excepthook = xplatform.guiexcepthook
    """
    try:
        import util
        import tempfile
        import traceback
        
        f = open(tempfile.mktemp(prefix='qtalchemy-error-', suffix='.txt'), "w")
        traceback.print_exception(type, value, tb, limit=None, file=f)

        util.xdg_open(f.name)
    except:
        import sys
        traceback.print_exception(type, value, tb, limit=None, file=sys.stderr)

class StdErrEater(object):
    """
    Instead of printing to stderr a GUI app may want to log the stderr stream 
    to a text file.  This is particularly relevant in cross-platform situations 
    because py2exe logs and messages the user at app exit.  We'd like to have a 
    friendlier approach.
    
    For windows, future possibilities include directing stderr lines to the 
    windows event log.
    
    A possible usage method is:

    >>> from qtalchemy import *
    >>> def exitApplication():
    ...     if isinstance(sys.stderr, xplatform.StdErrEater):
    ...         see = sys.stderr
    ...         if see.file:
    ...             fn = see.file.name
    ...             see.file.close()
    ...             see.file = None
    ... 
    ...             msgs = open(fn,"r").read()
    ...             if msgs != "":
    ...                 xplatform.xdg_open( fn )
    ...     
    >>> if xplatform.is_windows():
    ...     # We want a log of exceptions that have been sent to standard error.
    ...     # PyQt/PySide has a bad habit of not exposing these things to the user, but we absolutely must have a way to diagnose funny failures.
    ...     sys.stderr = xplatform.StdErrEater()
    ... 
    >>> # run the main body of the application
    >>> exitApplication()
    """
    
    # we ignore some errors we expect and don't care about
    ignore_these = [
        "does *not* support Decimal objects natively, and SQLAlchemy must convert from floating point"]

    def __init__(self):
        self.file = None
    
    def init_file(self):
        import tempfile
        import os
        import datetime
        x = tempfile.gettempdir()
        file_name = os.path.join(x,"camp_rental_error_{date}_{pid}.log".format(date=datetime.date.today().strftime("%Y%m%d"),pid=0))
        self.file = open(file_name,"a")
        self.file.write( """\
The following errors were logged during this program.  Please consider sending 
them to joel@kiwistrawberry.us for him to fix.  He might further inquire about 
whether you can reproduce the problem, but it's likely that the details below 
are sufficient to produce a fix.

""" )

    def write(self,s):
        for i in self.ignore_these:
            if s.find(i)>=0:
                return

        if self.file is None:
            self.init_file()

        if self.file is not None:
            self.file.write(s)



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
This module contains a python zip file extractor that works with py2exe.

It is not perfect, but it is a proof-of-concept that I'd rather not lose.
"""

from pkg_resources import resource_stream

resource_temp_dir = None

# http://old.nabble.com/resource_filename%28%29-breaks-cxfreeze-%28py2exe-%29-td1104344.html
# this is the motivation for the following function
def resource_to_tempfile(module, resource):
    """
    Read a pkg_resources stream and dump it to a temp file.
    Return the file name of the new temp file.

    TODO:  get smart about cacheing files and temp dir management!
    """
    import tempfile
    import os
    global resource_temp_dir

    if resource_temp_dir is None:
        resource_temp_dir = tempfile.mkdtemp()
    #print temp_dir,module,resource
    file_name = os.path.join(resource_temp_dir,module,resource)
    if not os.path.exists(file_name):
        my_stream = resource_stream(module,resource)
        if not os.path.exists(os.path.join(resource_temp_dir,module)):
            os.mkdir(os.path.join(resource_temp_dir,module))
        f = open(file_name,"wb")
        f.write(my_stream.read())
        f.close()
    return file_name



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
The PBDateEdit provides a date input box with more flexible entry.

It also supports null dates which is something QDateEdit does not.
"""

from PyQt4 import QtCore, QtGui
from button_edit import PBButtonEdit, Property
import datetime

import fuzzyparsers

from qtalchemy import fromQType, toQType

class DateValidator(QtGui.QValidator):
    def __init__(self,parent=None):
        QtGui.QValidator.__init__(self,parent)

    def fixup(self,input):
        if input == "":
            return ""
        try:
            date = fuzzyparsers.parse_date(input)
        except:
            return ""
        #input.replace(0,input.length(),date.strftime("%x"))
        print date.strftime("%x")
        return date.strftime("%x")

    def validate(self,input,pos):
        try:
            date = fuzzyparsers.parse_date(input)
            return QtGui.QValidator.Acceptable,input,pos
        except Exception,e:
            return QtGui.QValidator.Intermediate,input,pos
    

class PBDateEdit(PBButtonEdit):
    """
    PBDateEdit is a QLineEdit derivative that parses input strings into dates 
    with the fuzzyparsers python package.  A QCalendarWidget is available 
    by clicking a button to the right of the edit or pressing F4.
    """
    def __init__(self,parent=None):
        PBButtonEdit.__init__(self,parent)
        self.setValidator(DateValidator())
        self.button.setIcon(QtGui.QIcon(':/qtalchemy/widgets/view-calendar.ico'))
        self.editingFinished.connect(self.transform)
        x = self.sizePolicy()
        self.setSizePolicy(QtGui.QSizePolicy.Minimum,x.verticalPolicy())

    def minimumSizeHint(self):
        buttonWidth = self.style().pixelMetric(QtGui.QStyle.PM_ScrollBarExtent)
        x = PBButtonEdit.minimumSizeHint(self)
        x.setWidth(len(datetime.date.today().strftime("%x"))*9+buttonWidth)
        return x

    def sizeHint(self):
        buttonWidth = self.style().pixelMetric(QtGui.QStyle.PM_ScrollBarExtent)
        x = PBButtonEdit.sizeHint(self)
        x.setWidth(len(datetime.date.today().strftime("%x"))*9+buttonWidth)
        return x

    def date(self):
        x = self.text()
        if x == "":
            return None
        else:
            return toQType(fuzzyparsers.parse_date(x))

    def setDate(self,v):
        v = fromQType(v)
        if v is None:
            self.setText("")
        else:
            self.setText(v.strftime("%x"))

    date = Property("QDate", date, setDate)

    def transform(self):
        x = self.text()
        if x != "":
            self.setDate(toQType(fuzzyparsers.parse_date(x)))

    def date_selected(self,date):
        self.setDate(date)
        self.calendar.close()
        #self.calendar = None

    def buttonPress(self):
        PBButtonEdit.buttonPress(self)
        self.calendar = QtGui.QCalendarWidget()
        self.calendar.setWindowFlags(QtCore.Qt.Popup)
        if self.date is not None:
            self.calendar.setSelectedDate(self.date)
        self.calendar.activated.connect(self.date_selected)
        self.calendar.clicked.connect(self.date_selected)
        self.calendar.move(self.mapToGlobal(self.rect().bottomLeft()))
        self.calendar.show()
        self.calendar.setFocus(QtCore.Qt.PopupFocusReason)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup

import os
import dircache
import distutils
from distutils.core import Command
from distutils.command.build import build

def needsupdate(src, targ):
    return not os.path.exists(targ) or os.path.getmtime(src) > os.path.getmtime(targ)

class PySideUiBuild:
    def qrc(self, qrc_file, py_file):
        import subprocess
        rccprocess = subprocess.Popen(['pyside-rcc', qrc_file, '-py2', '-o', py_file])
        rccprocess.wait()

    def uic(self, ui_file, py_file):
        import subprocess
        rccprocess = subprocess.Popen(['pyside-uic', ui_file, '-o', py_file])
        rccprocess.wait()

class PyQt4UiBuild:
    def qrc(self, qrc_file, py_file):
        import subprocess
        rccprocess = subprocess.Popen(['pyrcc4', qrc_file, '-py2', '-o', py_file])
        rccprocess.wait()

    def uic(self, ui_file, py_file):
        from PyQt4 import uic
        fp = open(py_file, 'w')
        uic.compileUi(ui_file, fp)
        fp.close()

class QtUiBuild(Command, PyQt4UiBuild):
    description = "build Python modules from Qt Designer .ui files"

    user_options = []
    ui_files = []
    qrc_files = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def compile_ui(self, ui_file, py_file):
        if not needsupdate(ui_file, py_file):
            return
        print("compiling %s -> %s" % (ui_file, py_file))
        try:
            self.uic(ui_file, py_file)
        except Exception, e:
            raise distutils.errors.DistutilsExecError, 'Unable to compile user interface %s' % str(e)
            return
    
    def compile_qrc(self, qrc_file, py_file):
        if not needsupdate(qrc_file, py_file):
            return
        print("compiling %s -> %s" % (qrc_file, py_file))
        try:
            self.qrc(qrc_file, py_file)
        except Exception, e:
            raise distutils.errors.DistutilsExecError, 'Unable to compile resource file %s' % str(e)
            return

    def run(self):
        for f in self.ui_files:
            dir, basename = os.path.split(f)
            self.compile_ui(f, os.path.join(dir, "ui_"+basename.replace(".ui", ".py")))
        for f in self.qrc_files:
            dir, basename = os.path.split(f)
            self.compile_qrc(f, os.path.join(dir, basename.replace(".qrc", "_rc.py")))

QtUiBuild.ui_files = []
QtUiBuild.qrc_files = [os.path.join(dir, f) \
                for dir in ['qtalchemy/widgets', 'qtalchemy'] \
                for f in dircache.listdir(dir) if f.endswith('.qrc')]

class QtAlchemyBuild(build):
    sub_commands = [('build_ui', None)] + build.sub_commands

cmds = {
        'build' : QtAlchemyBuild,
        'build_ui' : QtUiBuild,
       }

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(name='qtalchemy',
      version='0.7.0',
      description='QtAlchemy is a framework for developing GUI database applications using SQLAlchemy and PyQt/PySide.',
      license='GPLv2+',
      author='Joel B. Mohler',
      author_email='joel@kiwistrawberry.us',
      long_description=read('README.txt'),
      url='https://bitbucket.org/jbmohler/qtalchemy/',
      packages=['qtalchemy',
                'qtalchemy/dialogs',
                'qtalchemy/widgets',
                'qtalchemy/xplatform',
                'qtalchemy/ext',
                'qtalchemy/ext/dataimport'],
      cmdclass = cmds,
      install_requires=['sqlalchemy', 'fuzzyparsers'],
      classifiers = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python :: 2",
        "Topic :: Database :: Front-Ends",
        "Operating System :: OS Independent"])



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################

from qtalchemy import *
from qtalchemy.widgets import *
from PyQt4 import QtCore, QtGui
from sqlalchemy.orm import Query

class QueryManager(ModelObject):
    """
    Abstract base class for QueryManager classes.
    """
    def objectConverter(self, obj):
        """
        Return the key value from the passed obj.  This function is passed 
        as the callable to the :class:`qtalchemy.QueryTableModel` which is in the :class:`QueryDataView`.
        """
        pass

    def fillQueryLayout(self, layout):
        """
        This function is called to give the manager a chance to create 
        the widgets which let the user input query selections.
        
        :param layout:  a QLayout object to receive the input elements.
        
        The helper functions :func:`qtalchemy.LayoutLayout` and :func:`qtalchemy.LayoutWidget` 
        are often helpful here.
        
        The return value is the widget that should be the focus proxy for the 
        enclosing :class:`QueryDataView`.
        """
        pass

    def requery(self):
        """
        This function uses the input objects in constructed in :func:`fillQueryLayout` to 
        construct a sqlalchemy.orm.Query object which will be associated with a 
        session and queried by the :class:`QueryDataView`.
        """
        pass



class BasicQueryManager(QueryManager):
    """
    This is the most basic useful override of :class:`QueryManager` possible.  
    It returns a query with the cols specified in the list of column names cols.
    
    :param cls: a sqlalchemy mapped class
    :param keyAttr: the primary key attribute name for cls
    :param cols: a list of attribute names desired in the view
    """
    def __init__(self, cls, keyAttr, cols):
        self.cls = cls
        self.keyAttr = keyAttr
        self.cols = cols
    
    def objectConverter(self, obj):
        return getattr(obj, self.keyAttr)

    def fillQueryLayout(self, layout):
        # no query options at all
        return None

    def requery(self):
        cols = [getattr(self.cls, c) for c in [self.keyAttr]+self.cols]
        return Query(cols)

class QueryDataView(QtGui.QWidget):
    """
    A QueryDataView displays a flexible interface element showing items in 
    a tabular view.  The main flexibility of QueryDataView arises from queryMgr 
    which is an instance of :class:`QueryManager`.  This object queryMgr 
    can add an arbitrary collection of input widgets for user query 
    parameters.  The queryMgr is responsible for returning a sqlalchemy Query 
    object back to the QueryDataView to display.
    
    :param parent:  a parent window (optionally pass as None)
    :param Session:  a session creator
    :param entityCls:  a command holding class from which the attributes in commandSets are taken
    :param commandSets:  a list of :class:`qtalchemy.CommandMenu` attributes in the entityCls
    :param queryMgr:  an instance of a :class:`QueryManager` class.

    >>> from sqlalchemy import Table, Column, String, Integer, MetaData, create_engine
    >>> from sqlalchemy.ext.declarative import declarative_base
    >>> 
    >>> from qtalchemy import *
    >>> from qtalchemy.ext.dataviews import *
    >>> from PyQt4 import QtCore, QtGui
    >>> 
    >>> metadata = MetaData()
    >>> Base = declarative_base(metadata=metadata, cls=ModelObject)
    >>> 
    >>> class People(Base):
    ...     __table__ = Table('people', metadata,
    ...                     Column('id', Integer, primary_key=True),
    ...                     Column('f_name', String(50), info={"label": "First Name"}),
    ...                     Column('l_name', String(50), info={"label": "Last Name"}),
    ...                     Column('title', String(10), info={"label": "Title"}))
    ... 
    >>> class PeopleCommands(object):
    ...     def __init__(self, Session, parent):
    ...         pass
    ...
    >>> engine = create_engine("sqlite://")
    >>> metadata.bind = engine
    >>> metadata.create_all()
    >>> Session = PBSessionMaker(bind=engine)
    >>> 
    >>> app = qtapp()
    >>> view = QueryDataView(None, Session, PeopleCommands, [], BasicQueryManager(People, "id", "f_name l_name title".split(' ')))
    """
    def __init__(self, parent, Session, entityCls, commandSets, queryMgr):
        QtGui.QWidget.__init__(self, parent)

        self.Session = Session
        self.entity = entityCls(Session, parent)
        self.queryManager = queryMgr

        self.setObjectName("{0}_{1}".format(self.entity.__class__.__name__, self.queryManager.__class__.__name__))

        layout = QtGui.QVBoxLayout()
        self.setLayout(layout)

        self.toolbar = LayoutWidget(layout, QtGui.QToolBar(self))
        self.toolbar.setFocusPolicy(QtCore.Qt.NoFocus)
        x = LayoutLayout(layout, QtGui.QHBoxLayout())
        self.setFocusProxy(self.queryManager.fillQueryLayout(x))
        self.refreshBtn = LayoutWidget(x, QtGui.QPushButton("Refresh"))
        self.table = LayoutWidget(layout, TableView())

        self.table.setModel(QueryTableModel(self.queryManager.requery(), ssrc=self.Session, objectConverter=self.queryManager.objectConverter), 
                            extensionId=suffixExtId(self, "Table"))

        self.refreshBtn.clicked.connect(self.refresh)

        self.bindings = [getattr(self.entity, c) for c in commandSets]
        self.bindings = [b.withView(self.table, bindDefault=True) for b in self.bindings]

        for i in range(len(self.bindings)):
            if i > 0:
                self.toolbar.addSeparator()
            self.bindings[i].fillToolbar(self.toolbar)
            self.bindings[i].preCommand.connect(self.preCommand)

        self.refresh()

    def keyPressEvent(self, e):
        if e.key() == QtCore.Qt.Key_Return or e.key() == QtCore.Qt.Key_Enter:
            self.refresh()
            e.accept()
        elif e.key() == QtCore.Qt.Key_Down:
            self.table.setFocus(QtCore.Qt.TabFocusReason)
            e.accept()
        else:
            QtGui.QWidget.keyPressEvent(self, e)

    def preCommand(self):
        pass
        #print "in preCommand"

    def refresh(self):
        self.table.model().query = self.queryManager.requery()
        self.table.model().reset_content_from_session()



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
from qtalchemy import *
from PyQt4 import QtCore, QtGui
from auth_settings import AuthSettings
from objectforms import MapperDialog, exception_message_box

class LoginDialog(MapperDialog):
    """
    >>> from qtalchemy.dialogs import LoginDialog
    >>> app = qtapp()
    >>> d=LoginDialog()
    >>> d.exec_()  #doctest: +SKIP
    0
    """
    def __init__(self,parent=None,init_session_maker=None,init_db=None):
        MapperDialog.__init__(self,parent)
        self.init_session_maker = init_session_maker
        self.init_db = init_db

        self.settings = AuthSettings()
        self.settings.loadSettings()

        self.setWindowTitle(self.tr("Log In"))

        main = QtGui.QVBoxLayout(self)
        user_grid = LayoutLayout(main,QtGui.QFormLayout())
        self.m_user = self.mapClass(AuthSettings)
        self.m_user.addBoundField(user_grid,"user")
        self.m_user.addBoundField(user_grid,"password")
        self.m_user.addBoundField(user_grid,"server_string")

        self.server_details = QtGui.QGroupBox("Server Details")
        main_ext = QtGui.QVBoxLayout(self.server_details)
        user_grid = LayoutLayout(main_ext,QtGui.QFormLayout())
        self.m_server = self.mapClass(AuthSettings)
        self.m_server.addBoundField(user_grid,"server")
        self.m_server.addBoundField(user_grid,"database")
        self.showing_details = False
        self.server_details.setVisible(self.showing_details)
        LayoutWidget(main,self.server_details)

        remember_layout = LayoutLayout(main,QtGui.QHBoxLayout())
        self.m_user.addBoundField(remember_layout,"remember_server")
        self.m_user.addBoundField(remember_layout,"remember_user")
        self.m_user.connect_instance(self.settings)
        self.m_server.connect_instance(self.settings)

        buttonbox = LayoutWidget(main,QtGui.QDialogButtonBox())
        buttonbox.addButton(QtGui.QDialogButtonBox.Ok)
        buttonbox.addButton(QtGui.QDialogButtonBox.Cancel)
        self.detailsBtn = ButtonBoxButton(buttonbox,QtGui.QPushButton("&Server"),QtGui.QDialogButtonBox.ActionRole)
        self.detailsBtn.clicked.connect(self.details_handler)
        buttonbox.accepted.connect(self.accept)
        buttonbox.rejected.connect(self.reject)

    def details_handler(self):
        self.showing_details = not self.showing_details
        self.server_details.setVisible(self.showing_details)
        self.setFixedHeight(self.sizeHint().height())
        self.setFixedWidth(self.sizeHint().width())

    def accept(self):
        self.submit()
        self.settings.saveSettings()

        self.connection = "%s://%s:%s@%s/%s" % (self.settings.server_type, self.settings.user, self.settings.password, self.settings.server, self.settings.database)
        try:
            #print self.connection
            if self.init_session_maker is not None:
                self.init_session_maker(self.connection)
            
            super(LoginDialog,self).accept()
        except Exception, e:
            exception_message_box(e,"There was an error initializing the data.",icon=QtGui.QMessageBox.Warning)

def auth_dialog(parent=None,init_session_maker=None,init_db=None):
    a = LoginDialog(parent,init_session_maker,init_db)
    a.show()
    return a.exec_()



# -*- coding: utf-8 -*-
##############################################################################
#       Copyright (C) 2010, Joel B. Mohler <joel@kiwistrawberry.us>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
##############################################################################
"""
The PBKeyEdit provides data sensitive editting for foreign key edits::

- search button
- context menus - for what? (look up referenced entity)
"""

from PyQt4 import QtCore, QtGui
from button_edit import PBButtonEdit

class PBKeyEdit(PBButtonEdit):
    """
    PBKeyEdit is a QLineEdit derivative that offers a button on the right to 
    search for rows from a database table.  PBKeyEdit is best used in the 
    InputYoke infrastructure with a DomainEntity derived class.
    """
    def __init__(self,parent=None):
        PBButtonEdit.__init__(self,parent)
        self.button.setIcon(QtGui.QIcon(':/qtalchemy/widgets/edit-find-6.ico'))



# -*- coding: utf-8 -*-
"""
We test for recursion asserts in UserAttr events
"""

from qtalchemy import UserAttr, AttrNumeric, ModelObject
import nose.tools

class InvoiceDetail(ModelObject):
    classEvents = ModelObject.Events()

    ordered = UserAttr(AttrNumeric(2),"Ordered")
    shipped = UserAttr(AttrNumeric(2),"Shipped")
    unit_price = UserAttr(AttrNumeric(2),"Unit Price")
    ordered_price = UserAttr(AttrNumeric(2),"Price (ordered)")
    shipped_price = UserAttr(AttrNumeric(2),"Price (shipped)")

    # internal
    unit_num = UserAttr(int,"unit numerator")
    unit_den = UserAttr(int,"unit denominator")

    def __repr__(self):
        return "{0.ordered} {0.shipped} {0.unit_price} {0.ordered_price} {0.shipped_price}\n{0.unit_num} {0.unit_den}".format(self)

    @unit_price.on_first_get
    def on_get_unit_price(self):
        if self.unit_num is None or self.unit_den is None:
            return 0.0
        return float(self.unit_num)/self.unit_den

    @classEvents.add("set", "unit_price")
    def on_change_unit_price(self, attr, oldvalue):
        if not self.is_setting("ordered_price") and not self.is_setting("shipped_price"):
            self.unit_num = int(self.unit_price * 100)
            self.unit_den = 100
            self.ordered_price = self.ordered * self.unit_price
            self.shipped_price = self.shipped * self.unit_price

    @shipped_price.on_first_get
    def on_get_shipped_price(self):
        if self.unit_num is None or self.unit_den is None:
            return 0.0
        return self.shipped * float(self.unit_num) / self.unit_den

    @classEvents.add("set", "shipped_price")
    def on_change_shipped_price(self, attr, oldvalue):
        if not self.is_setting("ordered_price") and not self.is_setting("unit_price"):
            self.unit_num = int(self.shipped_price * 100)
            self.unit_den = int(self.shipped * 100)
            self.unit_price = self.shipped_price / self.shipped
            self.ordered_price = self.ordered * self.unit_num / self.unit_den

    @ordered_price.on_first_get
    def on_get_ordered_price(self):
        if self.unit_num is None or self.unit_den is None:
            return 0.0
        return self.ordered * float(self.unit_num) / self.unit_den

    @classEvents.add("set", "ordered_price")
    def on_change_ordered_price(self, attr, oldvalue):
        if not self.is_setting("shipped_price") and not self.is_setting("unit_price"):
            self.unit_num = int(self.ordered_price * 100)
            self.unit_den = int(self.ordered * 100)
            self.unit_price = self.ordered_price / self.ordered
            self.shipped_price = self.shipped * self.unit_num / self.unit_den

def assign_unit_test():
    x = InvoiceDetail()
    x.ordered = 5.50
    x.shipped = 3.25
    x.unit_price = 12
    assert (x.ordered_price, x.shipped_price) == (66,39)

def assign_extended_3for1_test():
    x = InvoiceDetail()
    x.ordered = 6
    x.shipped = 3
    x.shipped_price = 100
    assert (x.unit_num, x.unit_den) == (10000,300)
    assert x.ordered_price == 200

def assign_extended_5for9_test():
    x = InvoiceDetail()
    x.ordered = 5
    x.shipped = 15
    x.ordered_price = .09
    assert (x.unit_num, x.unit_den) == (9,500)
    assert x.shipped_price*100 == 27

def assign_extended_test():
    x = InvoiceDetail()
    x.ordered = 5.50
    x.shipped = 3.25
    x.shipped_price = 50
    assert (x.unit_num, x.unit_den) == (5000,325)
    assert x.ordered_price*100 == 8462


class UnsafeInvoiceDetail(ModelObject):
    classEvents = ModelObject.Events()

    ordered = UserAttr(AttrNumeric(2),"Ordered")
    unit_price = UserAttr(AttrNumeric(2),"Unit Price")
    ordered_price = UserAttr(AttrNumeric(2),"Price (ordered)")
    
    # internal
    unit_num = UserAttr(int,"unit numerator")
    unit_den = UserAttr(int,"unit denominator")

    @unit_price.on_first_get
    def on_get_unit_price(self):
        if self.unit_num is None or self.unit_den is None:
            return 0.0
        return float(self.unit_num)/self.unit_den

    @classEvents.add("set", "unit_price")
    def on_change_unit_price(self, attr, oldvalue):
        self.unit_num = int(self.unit_price * 100)
        self.unit_den = 100
        self.ordered_price = self.ordered * self.unit_price

    @ordered_price.on_first_get
    def on_get_ordered_price(self):
        if self.unit_num is None or self.unit_den is None:
            return 0.0
        return self.ordered * float(self.unit_num) / self.unit_den

    @classEvents.add("set", "ordered_price")
    def on_change_ordered_price(self, attr, oldvalue):
        self.unit_num = int(self.ordered_price * 100)
        self.unit_den = int(self.ordered * 100)
        self.unit_price = self.ordered_price / self.ordered

@nose.tools.raises(AssertionError)
def unsafe_test():
    x = UnsafeInvoiceDetail()
    x.ordered = 2
    x.unit_price = 12  # this asserts because the class doesn't do recursion control
